# pylint: disable=missing-docstring,line-too-long

import unittest
from json import dumps
from typing import Any, Dict, Optional
from unittest.mock import MagicMock, patch

from soil.dictionary import dictionary

TEST_DICT = {
    "_id": "609bdfe9f5e2c907fa353bca99",
    "content": {"a": "b"},
    "created_at": 1620828137994,
    "language": "es",
    "name": "testdicaaddt",
}

TEST_CONTENT = {"a": "b"}


class MockResponse:  # pylint: disable=too-few-public-methods
    def __init__(self, text: str, status_code: int):
        self.text = text
        self.status_code = status_code

    # def text(self):
    #     return self._text


def mocked_requests_get(
    url: str, *, headers: Optional[Dict[str, str]] = None, timeout: int
) -> MockResponse:
    # pylint: disable=unused-argument
    name = url.split("/")[-2]
    if name == "testdicaaddt":
        return MockResponse(dumps(TEST_DICT), 200)
    return MockResponse("Not found", 404)


def mocked_requests_post(*_args: Any, **kwargs: Any) -> MockResponse:
    assert "json" in kwargs
    assert "language" in kwargs["json"]
    assert "name" in kwargs["json"]
    assert "content" in kwargs["json"]
    return MockResponse(dumps({"status": "created"}), 200)


# Our test case class
class TestDictionary(unittest.TestCase):
    @patch("soil.api.session.get", side_effect=mocked_requests_get)
    def test_get_dictionary(self, mock_get: MagicMock) -> None:
        response = dictionary("testdicaaddt", "ca")
        self.assertEqual(response, TEST_DICT)
        response = dictionary("testdicaaddt", "en")
        self.assertEqual(response, TEST_DICT)

        self.assertEqual(len(mock_get.call_args_list), 2)

    @patch("soil.api.session.patch", side_effect=mocked_requests_post)
    @patch("soil.api.session.post", side_effect=mocked_requests_post)
    @patch("soil.api.session.get", side_effect=mocked_requests_get)
    def test_create_existing_dictionary_no_lang(
        self, mock_get: MagicMock, mock_post: MagicMock, mock_patch: MagicMock
    ) -> None:
        response = dictionary("testdicaaddt", "ca", TEST_CONTENT)
        assert response == {"status": "created"}
        self.assertEqual(len(mock_get.call_args_list), 1)
        self.assertEqual(len(mock_post.call_args_list), 1)
        self.assertEqual(len(mock_patch.call_args_list), 0)

    @patch("soil.api.session.patch", side_effect=mocked_requests_post)
    @patch("soil.api.session.post", side_effect=mocked_requests_post)
    @patch("soil.api.session.get", side_effect=mocked_requests_get)
    def test_create_existing_dictionary_lang(
        self, mock_get: MagicMock, mock_post: MagicMock, mock_patch: MagicMock
    ) -> None:
        response = dictionary("testdicaaddt", "es", TEST_CONTENT)
        assert response == {"status": "created"}
        self.assertEqual(len(mock_get.call_args_list), 1)
        self.assertEqual(len(mock_post.call_args_list), 0)
        self.assertEqual(len(mock_patch.call_args_list), 1)

    @patch("soil.api.session.patch", side_effect=mocked_requests_post)
    @patch("soil.api.session.post", side_effect=mocked_requests_post)
    @patch("soil.api.session.get", side_effect=mocked_requests_get)
    def test_create_nonexisting_dictionary(
        self, mock_get: MagicMock, mock_post: MagicMock, mock_patch: MagicMock
    ) -> None:
        response = dictionary("testdicaaddtaaa", "es", TEST_CONTENT)
        assert response == {"status": "created"}
        self.assertEqual(len(mock_get.call_args_list), 1)
        self.assertEqual(len(mock_post.call_args_list), 1)
        self.assertEqual(len(mock_patch.call_args_list), 0)



"""
Soil package.
"""
from soil import connectors, errors, finder, types
from soil.alerts import alerts
from soil.alias import alias
from soil.data import data
from soil.decorator import decorator
from soil.dictionary import dictionary
from soil.logger import logger, logger_extra_kwarg, set_file_status

# pylint:disable=consider-using-from-import
from soil.modulify import modulify
from soil.task import task, task_wait

finder.upload_modules()

__all__ = [
    "modulify",
    "data_structures",
    "modules",
    "data",
    "alias",
    "logger",
    "connectors",
    "decorator",
    "task",
    "task_wait",
    "alerts",
    "dictionary",
    "errors",
    "types",
    "logger_extra_kwarg",
    "set_file_status",
]



"""Types used by soil library"""
# pylint:disable=unnecessary-ellipsis
from enum import Enum, StrEnum
from typing import Dict, List, Protocol, Self, Type, TypedDict

from soil.storage.base_storage import BaseStorage

Plan = List[Dict[str, str]]


class GetModule(TypedDict):
    """Type for GET modules/:moduleId"""

    is_package: bool
    public_api: List[str]
    package_type: str


class GetModuleHash(TypedDict):
    """Type for GET modules/"""

    name: str
    hash: str


class ExperimentStatuses(Enum):
    """Experiment Statuses"""

    WAITING = "WAITING"
    EXECUTING = "EXECUTING"
    DONE = "DONE"
    ERROR = "ERROR"


class Experiment(TypedDict):
    """Type for GET experiments/:experimentId"""

    _id: str
    app_id: str
    outputs: Dict[str, str]
    experiment_status: str


class Result(TypedDict):
    """Type for GET results/:resultId"""

    _id: str
    type: str


class SerializableDataStructure[Storage: BaseStorage, MetadataDict: TypedDict](
    Protocol
):
    """Data Strucutre base protocol"""

    @property
    def metadata(self: Self) -> MetadataDict:
        """Metadata of the DS protocol"""
        ...

    def serialize(self) -> Storage:
        """Serializes the DS."""
        ...

    @classmethod
    def deserialize(
        cls: Type[Self],
        storage: Storage,
        metadata: MetadataDict,
    ) -> "SerializableDataStructure[Storage, MetadataDict]":
        """Deserialize DS method."""
        ...


class TypeLog(StrEnum):
    """Types of Logs."""

    PROCESSED = "processed"
    NOT_CONSISTENT = "not_consistent"
    NOT_PROCESSED = "not_processed"



"""Module generates"""

import os.path
from hashlib import sha256
from itertools import islice
from pathlib import Path
from typing import Generator

from soil import api
from soil.types import GetModuleHash

FOLDER_BLACKLIST = [".vscode", ".venv", "test", ".git"]


BLOCK_SIZE = 1048576
MODULES_CHUNK = 25


def _hex_digest(data: bytes):
    file_hash = sha256()
    cursor = 0
    while cursor < len(data):
        file_hash.update(data[cursor : cursor + BLOCK_SIZE])
        cursor += BLOCK_SIZE
    return file_hash.hexdigest()


def _get_included_files() -> Generator[Path, None, None]:
    folders = [
        path
        for path in Path(".").glob("*")
        if path.is_dir() and path.name not in FOLDER_BLACKLIST
    ]
    for folder in folders:
        for selected_file in folder.rglob("*.py"):
            yield selected_file


def _filter_files(
    included_files: Generator[Path, None, None], file_hashes: list[GetModuleHash]
) -> Generator[tuple[str, str, bool], None, None]:
    file_hashes_dict = {module["name"]: module["hash"] for module in file_hashes}
    for file in included_files:
        code = file.read_text()
        code_hash = _hex_digest(bytes(code, encoding="utf-8"))
        module_name = str(file)[:-3].replace("/", ".")  # all are python files
        if file.stem == "__init__":
            module_name = module_name[: -len(".__init__")]
        if file_hashes_dict.get(module_name) != code_hash:
            yield (module_name, code, file.stem == "__init__")


def _upload_selected_modules(
    modules: Generator[tuple[str, str, bool], None, None]
) -> None:
    mods = list(islice(modules, MODULES_CHUNK))
    while len(mods) > 0:
        umods = []
        for module in mods:
            name, code, is_package = module
            umods.append({"name": name, "code": code, "is_package": is_package})
        api.upload_modules(modules=umods)
        mods = list(islice(modules, MODULES_CHUNK))


def _upload_modules() -> None:
    modules = api.get_modules()
    files_to_check = _get_included_files()
    modules_to_upload = _filter_files(files_to_check, modules)
    _upload_selected_modules(modules_to_upload)


def upload_modules() -> None:
    """Upload modules to soil when not in a test environment."""
    if os.environ.get("PY_ENV", "development") == "test":
        return
    _upload_modules()



"""Module that defines a Pipeline."""
# To prevent Pipeline not defined: https://stackoverflow.com/a/49872353/3481480
from __future__ import annotations

import copy
import datetime
import logging
from time import sleep
from typing import Any, Dict, List, Optional

from soil import api, errors
from soil.logger import logger as soil_logger
from soil.types import Experiment, ExperimentStatuses, Plan

# How much should wait between api calls
# Remember ES takes some time to index logs
SLEEP_TIME = 1

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Pipeline:
    """A Pipeline stores the transformations and dependencies to obtain certain results."""

    def __init__(self, plan: Optional[Plan] = None) -> None:  # noqa: D107
        self.plan = plan if plan is not None else []
        self.experiment: Optional[Experiment] = None
        self.finished = False

    def run(self) -> Dict[str, str]:
        """Run the Pipeline (blocking call until the experiment finishes)"""
        if self.finished and self.experiment:
            return self.experiment["outputs"]
        if self.experiment is None:
            experiment = api.create_experiment(self.plan)
            self.experiment = experiment
        status = api.get_experiment(experiment["_id"])["experiment_status"]
        start_date = datetime.datetime.now().astimezone().isoformat()
        while ExperimentStatuses(status) not in [
            ExperimentStatuses.DONE,
            ExperimentStatuses.ERROR,
        ]:
            sleep(SLEEP_TIME)
            logs = api.get_experiment_logs(experiment["_id"], start_date)
            _print_logs(logs)
            if len(logs) > 0:
                start_date = logs[0]["date"]
            status = api.get_experiment(experiment["_id"])["experiment_status"]
        sleep(SLEEP_TIME)
        logs = api.get_experiment_logs(experiment["_id"], start_date)
        _print_logs(logs)
        if ExperimentStatuses(status) == ExperimentStatuses.ERROR:
            raise errors.ExperimentError("Pipeline failed")
        logger.debug("experiment_done: %s", experiment["_id"])
        self.finished = True
        return experiment["outputs"]

    def add_transformation(self, transformation: Dict[str, str]) -> Pipeline:
        """Add a new transformation to the Pipeline, returns a new Pipeline
        containing the plan of the old Pipeline plus the transformation.
        """
        new_plan = [*self.plan, transformation]
        return Pipeline(plan=new_plan)

    @staticmethod
    def merge_pipelines(*pipelines: Pipeline) -> Pipeline:
        """Merges all the Pipelines passed into a new Pipeline that is returned."""
        merged_plan: Plan = sum([p.plan for p in pipelines], [])
        # Remove repeated
        merged_plan = [
            i for n, i in enumerate(merged_plan) if i not in merged_plan[n + 1 :]
        ]
        return Pipeline(plan=merged_plan)


def _print_logs(logs: List[Dict[str, Any]]) -> None:
    for log in logs[::-1]:
        level = getattr(logging, log["level"], logging.INFO)
        nlog = copy.copy(log)
        del nlog["message"]
        soil_logger.log(level, "%s - %s", nlog["date"], log["message"], extra=nlog)



"""Package that contains calls for soil alias data structures"""
from typing import cast

from soil.api import set_alias
from soil.data_structure import DataStructure
from soil.types import SerializableDataStructure


def alias(
    name: str,
    data_ref: DataStructure | SerializableDataStructure,
    roles: list[str] | None = None,
) -> None:
    """Set an alias to a data reference"""
    # Force to get an id if there is none yet.
    data_ref = cast(DataStructure, data_ref)
    if data_ref.id is None:
        data_ref.get_id()
    assert isinstance(data_ref.id, str)
    set_alias(name, data_ref.id, roles)



#  pylint:disable=unused-argument
"""Module that defines the @modulify decorator."""
import warnings
from functools import wraps
from types import GenericAlias
from typing import Any, Callable, TypeVar, cast, get_args

from soil.data_structure import DataStructure
from soil.decorator import decorator
from soil.pipeline import Pipeline
from soil.utils import generate_data_structure_ids, generate_transformation

_IterableGenericDataStructure = TypeVar(
    "_IterableGenericDataStructure", bound=["list", "tuple"]
)


def _omittable_parentheses(
    maybe_decorator: Callable[..., _IterableGenericDataStructure] | None = None,
    /,
    allow_partial: bool = False,
) -> Callable[..., _IterableGenericDataStructure]:
    # https://gist.github.com/oakkitten/03ca8f9c1113a7e32e32135e2cf5fef9
    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if allow_partial:
                if args and callable(args[-1]):
                    *args, fu = args  # pylint:disable=invalid-name
                    return func(*args, **kwargs)(fu)
            elif len(args) == 1 and callable(args[-1]) and not kwargs:
                return func()(args[0])
            return func(*args, **kwargs)

        return wrapper

    return _decorator if maybe_decorator is None else _decorator(maybe_decorator)


@_omittable_parentheses(allow_partial=True)
@decorator(depth=2)
def modulify(
    _func: Callable[..., _IterableGenericDataStructure] | None = None,
    *,
    output_types: Callable | None = None,
) -> Callable[..., _IterableGenericDataStructure]:
    """Decorates a function to mark it as a soil module."""

    def modulify_decorator(
        fn: Callable[..., _IterableGenericDataStructure]
    ) -> Callable[..., _IterableGenericDataStructure]:
        fnn = fn if _func is None else _func
        num_outputs = fnn.__annotations__.get("return", None)

        mod_name = fnn.__module__ + "." + fnn.__name__

        def decorated(
            *inputs: DataStructure, **kwargs: Any
        ) -> _IterableGenericDataStructure:
            num_outputs_in = None
            if output_types is not None:
                # WARNING this could fail if output_types() checks the type of *inputs
                num_outputs_in = len(output_types(*inputs, **kwargs))
                warnings.warn(
                    "output_types is deprecated use annotations on function return instead.",
                    DeprecationWarning,
                )
            if (
                isinstance(num_outputs, GenericAlias)
                and num_outputs.__origin__ == tuple
            ):
                num_outputs_in = len(get_args(num_outputs))
            if num_outputs_in is None:
                raise ValueError(
                    "@modulify: "
                    + mod_name
                    + " missing return annotation of decorated function "
                    "or annotation not instance of tuple."
                )

            input_pipes = [
                input.pipeline for input in inputs if input.pipeline is not None
            ]
            new_pipeline = Pipeline.merge_pipelines(*input_pipes)
            output_ids = generate_data_structure_ids(mod_name, num_outputs_in)
            input_ids = [d.id if d.id is not None else d.sym_id for d in inputs]
            for i in input_ids:
                assert isinstance(i, str)
            transformation = generate_transformation(
                input_ids=cast(list[str], input_ids),
                output_ids=output_ids,
                fn_name=mod_name,
                args=kwargs,
            )
            new_pipeline = new_pipeline.add_transformation(transformation)
            outputs = [
                DataStructure(sym_id=sym_id, pipeline=new_pipeline)
                for sym_id in output_ids
            ]
            return outputs

        return decorated

    if _func is None:
        return modulify_decorator
    return modulify_decorator(_func)



"""Package that contains calls to the SOIL's REST API"""
import json
import logging
from typing import Any, Dict, List, Optional, TypedDict
from urllib.parse import urlencode

import requests
from requests import Response
from requests.adapters import HTTPAdapter, Retry

from soil import errors
from soil.configuration import GLOBAL_CONFIG
from soil.types import Experiment, GetModule, GetModuleHash, Plan, Result

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

API_ROOT = f"{GLOBAL_CONFIG.host!s}/v2/"


class _SoilSession(requests.Session):
    """Creates a Soil session that will refresh the token when necessary."""

    def _add_header(self) -> None:
        self.headers.update({"Authorization": f"Bearer {GLOBAL_CONFIG.token!s}"})

    def get(self, *args, **kwargs) -> Response:
        self._add_header()
        return super().get(*args, **kwargs)

    def post(self, *args, **kwargs) -> Response:
        self._add_header()
        return super().post(*args, **kwargs)

    def put(self, *args, **kwargs) -> Response:
        self._add_header()
        return super().put(*args, **kwargs)

    def patch(self, *args, **kwargs) -> Response:
        self._add_header()
        return super().patch(*args, **kwargs)


def _setup_requests_session() -> _SoilSession:
    """Creates a requests session object to be reused in the API calls."""
    _session = _SoilSession()
    _session.headers.update(
        {
            "Content-Type": "application/json",
        }
    )
    # Retries at 2, 4, 8, 16, 32 seconds
    retries = Retry(
        total=5, backoff_factor=2, status_forcelist=[500, 501, 502, 503, 504]
    )
    adapter = HTTPAdapter(max_retries=retries)
    _session.mount("http://", adapter)
    _session.mount("https://", adapter)
    return _session


session = _setup_requests_session()


def upload_data(dtype: str, data: Any, metadata: Any) -> Result:
    """Upload data to the cloud as a new dataset."""
    logger.debug("upload_data:%s")
    url = API_ROOT + "results/"
    body = {"type": dtype, "data": data, "metadata": metadata}
    response = session.post(url, json=body, timeout=30)
    if response.status_code != 201:
        raise errors.DataNotUploaded(response.text)
    return json.loads(response.text)


def get_result(result_id: str) -> Dict[str, Any]:
    """Get the result data"""
    logger.debug("get_result:%s", result_id)
    url = API_ROOT + "results/" + result_id + "/"
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.DataNotFound(response.text)
    return json.loads(response.text)


def get_result_data(
    result_id: str, query: Optional[Dict[str, str]] = None
) -> Dict[str, Any]:
    """Get the result data"""
    if query is None:
        query = {}
    logger.debug("get_result:%s %s", result_id, str(query))
    url = API_ROOT + "results/" + result_id + "/data/?" + urlencode(query)
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.DataNotFound(response.text)
    return json.loads(response.text)


def export_result(
    result_id: str, file_path: str, query: Optional[Dict[str, str]] = None
) -> None:
    """Export result and saves it to a file"""
    if query is None:
        query = {}
    logger.debug("export_result:%s %s", result_id, str(query))
    url = API_ROOT + "results/" + result_id + "/export/?" + urlencode(query)
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.DataNotFound(response.text)
    with open(file_path, "wb") as fd:
        fd.write(response.content)


class _SoilUploadModule(TypedDict):
    name: str
    code: str
    is_package: bool


def upload_modules(modules: list[_SoilUploadModule]) -> None:
    """Uploads a chunk of modules"""
    logger.debug("upload_modules:%s", modules)
    url = API_ROOT + "modules/"
    data = {"modules": modules}
    response = session.post(url, json=data, timeout=30)
    if response.status_code != 200:
        raise errors.ModuleNotUploaded(response.text)


def upload_module(module_name: str, code: str, is_package: bool) -> None:
    """Uploads a module"""
    logger.debug("upload_module:%s", module_name)
    url = API_ROOT + "modules/"
    data = {"name": module_name, "code": code, "is_package": is_package}
    response = session.post(url, json=data, timeout=30)
    if response.status_code != 200:
        raise errors.ModuleNotUploaded(response.text)


def get_module(full_name: str) -> GetModule:
    """Downloads a module"""
    logger.debug("get_module:%s", full_name)
    url = API_ROOT + "modules/" + full_name + "/"
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.ModuleNotFound(response.text)
    return json.loads(response.text)


def get_modules() -> list[GetModuleHash]:
    """Gets a list"""
    logger.debug("get_module:%s")
    url = f"{API_ROOT}modules/"
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.ModuleNotFound(response.text)
    return json.loads(response.text)["result"]


def set_alias(alias: str, result_id: str, roles: Optional[List[str]] = None) -> None:
    """Sets an alias for a result. Updates a previous one with the same name."""
    logger.debug("set_alias: %s = %s", alias, result_id)
    obj = {"name": alias, "state": {"alias": alias, "result_id": result_id}}
    if roles is not None:
        obj["roles"] = roles
    try:
        old_alias_id = get_alias(alias)["_id"]
        url = API_ROOT + "states/" + old_alias_id + "/"
        response = session.patch(url, json=obj, timeout=30)
    except errors.DataNotFound:
        url = API_ROOT + "states/"
        response = session.post(url, json=obj, timeout=30)
    if response.status_code != 200:
        raise errors.DataNotUploaded(f"Failed to create alias {alias}: {response.text}")

    if roles is not None:
        url = API_ROOT + "results/" + result_id + "/"
        response = session.patch(url, json={"roles": roles}, timeout=30)
        if response.status_code != 200:
            raise errors.DataNotUploaded(
                f"Failed to patch result {result_id}: {response.text}"
            )


def get_alias(alias: str) -> Dict[str, Any]:
    """Gets an alias"""
    logger.debug("get_alias: %s", alias)
    url = API_ROOT + "states/?name=" + alias + "&with-roles"
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.ObjectNotFound("Error getting alias:" + response.text)
    aliases = json.loads(response.text)
    if len(aliases) == 0:
        raise errors.DataNotFound("Alias not found")
    alias_id = aliases[0]["_id"]
    url = API_ROOT + "states/" + alias_id + "/"
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.ObjectNotFound("Error getting alias with id " + response.text)
    return json.loads(response.text)


def create_experiment(plan: Plan) -> Experiment:
    """Runs an experiment in SOIL"""
    logger.debug("create_experiment: %s", str(plan))
    url = API_ROOT + "experiments/"
    experiment = {"name": "", "description": "", "plan": plan}
    response = session.post(url, json={"experiment": experiment}, timeout=30)
    if response.status_code != 200:
        raise ValueError("Error creating the experiment:" + response.text)
    return json.loads(response.text)["experiment"]


def get_experiment(experiment_id: str) -> Experiment:
    """Runs an experiment from SOIL"""
    logger.debug("get_experiment: %s", experiment_id)
    url = API_ROOT + "experiments/" + experiment_id + "/"
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise ValueError(
            "Error getting the experiment " + experiment_id + "\n" + response.text
        )
    return json.loads(response.text)


def get_experiment_logs(experiment_id: str, start_date: str) -> Any:
    """Gets logs from a SOIL experiment"""
    logger.debug("get_experiment_id: %s since %s", experiment_id, start_date)
    url = (
        API_ROOT
        + "experiments/"
        + experiment_id
        + "/logs/?"
        + urlencode({"start_date": start_date})
    )
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise ValueError(
            "Error getting the experiment " + experiment_id + "\n" + response.text
        )
    return json.loads(response.text)


def create_event(key: str, value: Any) -> Any:
    """Saves an event in soil"""
    url = API_ROOT + "alerts/events/"
    response = session.post(url, json={"key": key, "value": value}, timeout=30)
    if response.status_code != 201:
        raise errors.EventNotUploaded("Error saving the event:" + response.text)
    return json.loads(response.text)


def create_alert(alert: Dict) -> Any:
    """Creates an alert"""
    url = API_ROOT + "alerts/alerts/"
    response = session.post(url, json=alert, timeout=30)
    if response.status_code != 201:
        raise errors.AlertNotUploaded("Error creating the alert:" + response.text)
    return json.loads(response.text)


def get_dictionary(name: str, language: str) -> Dict[str, Any]:
    """Get the a dictionary"""
    url = API_ROOT + "dictionaries/" + name + "/?language=" + language
    response = session.get(url, timeout=30)
    if response.status_code != 200:
        raise errors.DictionaryNotFound("Error getting dictionary:" + response.text)
    return json.loads(response.text)


def create_dictionary(name: str, language: str, content: Dict) -> Dict[str, Any]:
    """Create a dictionary or update it"""
    logger.debug("create_dictionary: %s in language: %s", str(name), str(language))

    try:
        dictionary = get_dictionary(name, language)
        dictionary_exists = dictionary["language"] == language
    except errors.DictionaryNotFound:
        dictionary_exists = False

    data = {
        "name": name,
        "language": language,
        "content": content,
    }

    if dictionary_exists:
        url = f"{API_ROOT}dictionaries/{name}/"
        response = session.patch(url, json=data, timeout=30)
    else:
        url = f"{API_ROOT}dictionaries/"
        response = session.post(url, json=data, timeout=30)

    if response.status_code != 200:
        raise errors.DictionaryNotUploaded(
            f"Error creating/updating the dictionary {name} in language {language}: {response.text}"
        )
    return json.loads(response.text)



# pylint: disable=missing-docstring,line-too-long, unused-argument, redefined-outer-name
# type: ignore

import unittest
from unittest.mock import patch

from pandas import Timestamp

import soil
from soil import decorator, modulify, task
from soil.data_structure import DataStructure


@modulify(output_types=lambda *input_types, **args: [input_types[0]])
def train(*data, module=None, model_params=None, **kwargs):
    (res,) = task(module)(*data, **model_params)
    return [res]


@decorator(depth=3)
def trainable(module):
    def decorator(**model_params):
        def train_fn(**train_params):
            def inner_module(*data):
                return train(
                    *data, module=module, model_params=model_params, **train_params
                )

            return inner_module

        return train_fn

    return decorator


@trainable
@modulify(output_types=lambda *input_types, **args: [input_types[0]])
def my_model(*data, **my_model_params):
    if "p3" in my_model_params:
        return task(my_model_params["p3"](cross_validation=True))(*data)
    return data


@modulify(output_types=lambda *input_types, **args: [input_types[0]])
def my_model2(*data, **my_model_params):
    return data


@modulify
def my_model3(*data, **my_model_params) -> tuple[DataStructure]:
    return data


def test_fn1():
    pass


def test_fn2():
    pass


def mock_soil_data_fn(data):
    return DataStructure("test_id")


class TestUtils(unittest.TestCase):
    @patch("soil.data")
    def test_build_arguments_1(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": [1, 2, 3]}])
        (res,) = my_model(test_param=1)(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected_args = {
            "module": {
                "__soil_arg_type": "decorated",
                "module": "soil.modulify",
                "name": "modulify",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
            },
            "model_params": {"test_param": 1},
            "cross_validation": True,
        }
        assert first["module"] == "unit.soil.test_utils.train"
        assert args == expected_args

    @patch("soil.data")
    def test_build_arguments_2(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=[test_fn1, test_fn2])(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": [
                    {
                        "__soil_arg_type": "function",
                        "module": "unit.soil.test_utils",
                        "name": "test_fn1",
                    },
                    {
                        "__soil_arg_type": "function",
                        "module": "unit.soil.test_utils",
                        "name": "test_fn2",
                    },
                ]
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_3(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=my_model2)(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [],
                            "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                        },
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "my_model2",
                                }
                            ],
                            "kwargs": {},
                        },
                    ],
                    "module": "soil.modulify",
                    "name": "modulify",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_4(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=trainable(my_model2))(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model2",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_5(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=my_model)(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        }
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_6(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=my_model(f1=2, t1=test_fn1))(
            cross_validation=True
        )(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "test_fn1",
                                },
                            },
                        },
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_7(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=my_model(f1=2, t1=test_fn1)(another=1))(
            cross_validation=True
        )(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "test_fn1",
                                },
                            },
                        },
                        {"args": [], "kwargs": {"another": 1}},
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_8(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(
            test_param=trainable(my_model2)(f1=2, t1=test_fn1)(another=1)
        )(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model2",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "function",
                                    "module": "unit.soil.test_utils",
                                    "name": "test_fn1",
                                },
                            },
                        },
                        {"args": [], "kwargs": {"another": 1}},
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_9(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(
            test_param=trainable(my_model2)(f1=2, t1=my_model(test=1))(another=1)
        )(cross_validation=True)(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model2",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "decorated",
                                                    "calls": [
                                                        {
                                                            "args": [],
                                                            "kwargs": {
                                                                "output_types": {
                                                                    "__soil_arg_type": "lambda"
                                                                }
                                                            },
                                                        },
                                                        {
                                                            "args": [
                                                                {
                                                                    "__soil_arg_type": "function",
                                                                    "module": "unit.soil.test_utils",
                                                                    "name": "my_model",
                                                                }
                                                            ],
                                                            "kwargs": {},
                                                        },
                                                    ],
                                                    "module": "soil.modulify",
                                                    "name": "modulify",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                        {"args": [], "kwargs": {"test": 1}},
                                    ],
                                    "module": "unit.soil.test_utils",
                                    "name": "trainable",
                                },
                            },
                        },
                        {"args": [], "kwargs": {"another": 1}},
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_10(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        (res,) = my_model(test_param=my_model(f1=2, t1=my_model(test=1))(another=1))(
            cross_validation=True
        )(data)
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "cross_validation": True,
            "model_params": {
                "test_param": {
                    "__soil_arg_type": "decorated",
                    "calls": [
                        {
                            "args": [
                                {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [],
                                            "kwargs": {
                                                "output_types": {
                                                    "__soil_arg_type": "lambda"
                                                }
                                            },
                                        },
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "function",
                                                    "module": "unit.soil.test_utils",
                                                    "name": "my_model",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                    ],
                                    "module": "soil.modulify",
                                    "name": "modulify",
                                }
                            ],
                            "kwargs": {},
                        },
                        {
                            "args": [],
                            "kwargs": {
                                "f1": 2,
                                "t1": {
                                    "__soil_arg_type": "decorated",
                                    "calls": [
                                        {
                                            "args": [
                                                {
                                                    "__soil_arg_type": "decorated",
                                                    "calls": [
                                                        {
                                                            "args": [],
                                                            "kwargs": {
                                                                "output_types": {
                                                                    "__soil_arg_type": "lambda"
                                                                }
                                                            },
                                                        },
                                                        {
                                                            "args": [
                                                                {
                                                                    "__soil_arg_type": "function",
                                                                    "module": "unit.soil.test_utils",
                                                                    "name": "my_model",
                                                                }
                                                            ],
                                                            "kwargs": {},
                                                        },
                                                    ],
                                                    "module": "soil.modulify",
                                                    "name": "modulify",
                                                }
                                            ],
                                            "kwargs": {},
                                        },
                                        {"args": [], "kwargs": {"test": 1}},
                                    ],
                                    "module": "unit.soil.test_utils",
                                    "name": "trainable",
                                },
                            },
                        },
                        {"args": [], "kwargs": {"another": 1}},
                    ],
                    "module": "unit.soil.test_utils",
                    "name": "trainable",
                }
            },
            "module": {
                "__soil_arg_type": "decorated",
                "calls": [
                    {
                        "args": [],
                        "kwargs": {"output_types": {"__soil_arg_type": "lambda"}},
                    },
                    {
                        "args": [
                            {
                                "__soil_arg_type": "function",
                                "module": "unit.soil.test_utils",
                                "name": "my_model",
                            }
                        ],
                        "kwargs": {},
                    },
                ],
                "module": "soil.modulify",
                "name": "modulify",
            },
        }
        assert args == expected

    @patch("soil.data")
    def test_build_arguments_11(self, mock_soil_data):
        mock_soil_data.side_effect = mock_soil_data_fn
        data = soil.data([{"test": "test"}])
        param = Timestamp("2020-02-02 02:02:02+01:00").tz_convert("Europe/Madrid")
        param1 = Timestamp("2020-02-02 02:02:02+02:00")
        param2 = Timestamp("2020-02-02 02:02:02")
        (res,) = my_model3(
            data, test_param=param, test_param_1=param1, test_param_2=param2
        )
        plan = res.pipeline.plan
        first = plan[0]
        args = first["args"]
        expected = {
            "test_param": {
                "_object": "Timestamp",
                "_constructor_params": {
                    "positional_args": ["2020-02-02T02:02:02+01:00"],
                    "keywords_args": {"tz": "Europe/Madrid"},
                },
            },
            "test_param_1": {
                "_object": "Timestamp",
                "_constructor_params": {
                    "positional_args": ["2020-02-02T02:02:02+02:00"],
                    "keywords_args": {},
                },
            },
            "test_param_2": {
                "_object": "Timestamp",
                "_constructor_params": {
                    "positional_args": ["2020-02-02T02:02:02"],
                    "keywords_args": {},
                },
            },
        }
        self.assertDictEqual(args, expected)
